# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"getcatdict": "00_utils.ipynb",
         "store_bson": "00_utils.ipynb",
         "load_bson": "00_utils.ipynb",
         "generate_files_embedprojector_for_comparision": "00_utils.ipynb",
         "JSONizerWithBool": "01_extract.ipynb",
         "get_metadict_skeleton": "02_transfer.ipynb"}

modules = ["utils.py",
           "extract.py",
           "transfer.py"]

doc_url = "https://manikyabard.github.io/transfertab/"

git_url = "https://github.com/manikyabard/transfertab/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/02_transfer.ipynb (unless otherwise specified).

__all__ = ['get_metadict_skeleton', 'transferembeds_']

# Cell
import pathlib
import torch
import torch.nn as nn
import json
import pandas as pd
from functools import partial
from fastcore.foundation import *
from fastcore.dispatch import *
from nbdev.showdoc import *
from .utils import *
from .extract import *

# Cell
def get_metadict_skeleton(df: pd.DataFrame, *, catcols=None, path=None):
    catdict = getcatdict(df, catcols)
    metadict = {}
    for (cat, classes) in catdict.items():
        metadict[cat] = {'mapped_cat': '', 'classes_info': {clas: [] for clas in classes}}
    if path != None:
        with open(path, 'w') as fp:
            json.dump(metadict, fp)
    return metadict

# Cell
@typedispatch
def transferembeds_(
        dest_embeds: nn.Module,
        src_embeds: nn.Module,
        # /,
        metatransfer,
        transfer_cats,
        # *,
        newcatcols,
        oldcatcols,
        oldcatdict,
        newcatdict,
        aggfn = partial(torch.mean, dim=0)):
    '''
        Transfers embeddings from `src_embeds` to `dest_embeds`,
        with the help of collections containing various metadata.
    '''
    src_state_dict = L(src_embeds.state_dict().items())
    dest_state_dict = L(dest_embeds.state_dict().items())
    for newcat in transfer_cats:
        newidx = newcatcols.index(newcat)
        oldidx = oldcatcols.index(metatransfer[newcat]["mapped_cat"])
        new_ps = torch.zeros(src_state_dict[oldidx][1].shape[1], 0)
        for newclass in newcatdict[newcat]:
            classidxs = L(oldcatdict[oldcatcols[oldidx]]).argwhere(lambda x: x in metatransfer[newcat]["classes_info"][newclass])
            if len(classidxs) == 0:
                classidxs =  list(range(len(oldcatdict[oldcatcols[oldidx]])))
            ps = torch.unsqueeze(aggfn(torch.index_select(src_state_dict[oldidx][1], 0, torch.LongTensor(classidxs))), -1)
            new_ps = torch.cat((new_ps, ps), dim=1)
        dest_embeds.state_dict()[dest_state_dict[newidx][0]].copy_(new_ps.T)

@typedispatch
def transferembeds_(
        dest_embeds: nn.Module,
        src_embeds: dict,
        metatransfer,
        transfer_cats,
        # *,
        newcatcols,
        oldcatcols,
        newcatdict,
        aggfn = partial(torch.mean, dim=0)):
    dest_state_dict = L(dest_embeds.state_dict().items())
    for newcat in transfer_cats:
        newidx = newcatcols.index(newcat)
        oldcatname = metatransfer[newcat]['mapped_cat']
        new_ps = torch.zeros(torch.tensor(src_embeds[oldcatname]['embeddings']).shape[1], 0)
        for newclass in newcatdict[newcat]:
            classidxs = L(src_embeds[oldcatname]['classes']).argwhere(lambda x: x in metatransfer[newcat]["classes_info"][newclass])
            if len(classidxs) == 0:
                classidxs = list(range(len(src_embeds[oldcatname]['classes'])))
            ps = torch.unsqueeze(aggfn(torch.index_select(torch.tensor(src_embeds[oldcatname]['embeddings']), 0, torch.LongTensor(classidxs))), -1)
            new_ps = torch.cat((new_ps, ps), dim=1)
        dest_embeds.state_dict()[dest_state_dict[newidx][0]].copy_(new_ps.T)


@typedispatch
def transferembeds_(
        dest_embeds: nn.Module,
        src_embeds: (pathlib.PosixPath, str),
        metatransfer,
        transfer_cats,
        # *,
        kind = "bson",
        **kwargs):
    if kind == "json":
        with open(src_embeds, 'r') as fp:
            src_embeds = json.loads(fp.read())
    else:
        src_embeds = load_bson(src_embeds)
    transferembeds_(dest_embeds, src_embeds, metatransfer, transfer_cats, **kwargs);


# Cell
#nbdev_comment _all_ = ['transferembeds_']


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/00_utils.ipynb (unless otherwise specified).

__all__ = ['getcatdict', 'store_bson', 'load_bson', 'generate_files_embedprojector_for_comparision',
           '_load_extractembeds_from_dir', '_check_embdict_validity']

# Cell
import bson
import json
from pathlib import Path

# Cell
def getcatdict(df, catcols=None, add_na=False):
    if catcols == None:
        columns = list(df.columns)
        catcols = list(map(lambda arg: columns[arg[0]], filter(lambda arg: arg[1] == 'object', enumerate(df.dtypes))))
    catdict = {}
    for cat in catcols:
        catdict[cat] = ["nan"] + list(df[cat].unique()) if add_na else list(df[cat].unique())
    return catdict

# Cell
def store_bson(path, data):
    bdata = bson.dumps(data)
    with open(path, "wb") as fp:
        fp.write(bdata)

def load_bson(path):
    with open(path, "rb") as fp:
        bdata = fp.read()
    return bson.loads(bdata)

# Cell
def _load_extractembeds_from_dir(path):
    path = Path(path)
    json_file = {}
    if path.is_dir():
        for json_f in path.glob("*.json"):
            f = open(json_f, 'rb')
            file = json.load(f)
            for i in file.keys():
                json_file[i] = file[i]
        for bson_f in path.glob("*.bson"):
            file = load_bson(bson_f)
            for i in file.keys():
                json_file[i] = file[i]
    return json_file

# Cell
def _check_embdict_validity(embdict):
    for catcol, info in embdict.items():
        if set(list(info.keys())) == set(('classes', 'embeddings')) :
            class_len = len(info['classes'])
            embedding_len =  len(info['embeddings'])
            assert class_len == embedding_len, f"Class Size of {class_len} and Embedding size of {embedding_len} does not match for {catcol}"

# Cell
def generate_files_embedprojector_for_comparision(model1_path, model2_path, dir_path):
    model1_embedsdict = load_bson(model1_path)
    model2_embedsdict = load_bson(model2_path)
    for ((catcol, info1), (_, info2)) in zip(model1_embedsdict.items(), model2_embedsdict.items()):
        model1_data = [[clas + '_Model1'] for clas in info1['classes']]
        model2_data = [[clas + '_Model2'] for clas in info2['classes']]
        with open(dir_path + '/' + catcol + '_meta.tsv', 'w') as tsvfile:
            writer = csv.writer(tsvfile, delimiter='\n')
            writer.writerows(model1_data)
            writer.writerows(model2_data)
        with open(dir_path + '/' + catcol + '_embeds.tsv', 'w') as tsvfile:
            writer = csv.writer(tsvfile, delimiter='\t')
            writer.writerows(info1['embeddings'])
            writer.writerows(info2['embeddings'])

# Cell
#nbdev_comment _all_ = ['_load_extractembeds_from_dir', '_check_embdict_validity']


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/01_extract.ipynb (unless otherwise specified).

__all__ = ['JSONizerWithBool', 'extractembeds']

# Cell
from copy import deepcopy
import json
import torch.nn as nn
import pandas as pd
from fastcore.dispatch import *
from .utils import *

# Cell
class JSONizerWithBool(json.JSONEncoder):
    def default(self, obj):
        return super().encode(bool(obj)) \
            if isinstance(obj, np.bool_) \
            else super().default(obj)

# Cell
def _catdict2embedsdictstruct(catdict):
    embedsdict = {}
    for cat, classes in catdict.items():
        embedsdict[cat] = {}
        embedsdict[cat]["classes"] = classes
    return embedsdict

# Cell
@typedispatch
def extractembeds(model: nn.Module, df: pd.DataFrame, *, transfercats, allcats, path=None, kind="bson"):
    catdict = getcatdict(df, transfercats)
    return extractembeds(model, catdict, transfercats=transfercats, allcats=allcats, path=path, kind=kind)


@typedispatch
def extractembeds(model: nn.Module, catdict: dict, *, transfercats, allcats, path=None, kind="bson"):
    '''
    Extracts embedding weights from `model`, which can be further transferred to other models.

    model: Any pytorch model, containing the embedding layers.
    catdict: A dictionary with category as key, and classes as value.
    transfercats: Names of categories to be transferred.
    allcats: Names of all categories corresponding to the embedding layers in model.
    path: Path for the json to be stored.
    '''
    embedsdict = _catdict2embedsdictstruct(catdict)
    model_dict = list(model.state_dict().items())
    for i, cat in enumerate(transfercats):
        classes = catdict[cat]
        catidx = allcats.index(cat)
        assert (model_dict[catidx][1].shape[0] == len(classes)), \
            (f"embeddings dimension {model_dict[catidx][1].shape[0]} !="
            f"num of classes {len(classes)} for vairable {cat}. Embeddings should have"
            f"same number of classes. Something might have gone wrong.")
        embedsdict[cat]["embeddings"] = model_dict[catidx][1].numpy().tolist()
    if (path != None):
        with open(path, 'w') as fp:
            json.dump(embedsdict, fp, cls = JSONizerWithBool) if kind == "json" else store_bson(path, embedsdict)
    return embedsdict

# Cell
#nbdev_comment _all_ = ['extractembeds']